@@ -37,6 +37,7 @@
 **
 ****************************************************************************/
 
+#include "qlowenergycharacteristicdata.h"
 #include "qbluetoothaddress.h"
 #include "osxbtutility_p.h"
 #include "qbluetoothuuid.h"
@@ -70,6 +71,10 @@ Q_LOGGING_CATEGORY(QT_BT_OSX, "qt.bluetooth.ios")
 
 namespace OSXBluetooth {
 
+const int defaultLEScanTimeoutMS = 25000;
+// We use it only on iOS for now:
+const int maxValueLength = 512;
+
 QString qt_address(NSString *address)
 {
     if (address && address.length) {
@@ -314,6 +319,19 @@ ObjCStrongReference<NSData> data_from_bytearray(const QByteArray & qtData)
     return result;
 }
 
+ObjCStrongReference<NSMutableData> mutable_data_from_bytearray(const QByteArray &qtData)
+{
+    using MutableData = ObjCStrongReference<NSMutableData>;
+
+    if (!qtData.size())
+        return MutableData([[NSMutableData alloc] init], false);
+
+    MutableData result([[NSMutableData alloc] initWithLength:qtData.size()], false);
+    [result replaceBytesInRange:NSMakeRange(0, qtData.size())
+                      withBytes:qtData.constData()];
+    return result;
+}
+
 // A small RAII class for a dispatch queue.
 class SerialDispatchQueue
 {
@@ -350,11 +368,6 @@ dispatch_queue_t qt_LE_queue()
     return leQueue.data();
 }
 
-unsigned qt_LE_deviceInquiryLength()
-{
-    return 10;
-}
-
 }
 
 QT_END_NAMESPACE
