@@ -1,8 +1,8 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Lauri Laanmets (Proekspert AS) <lauri.laanmets@eesti.ee>
-** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 Lauri Laanmets (Proekspert AS) <lauri.laanmets@eesti.ee>
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtBluetooth module of the Qt Toolkit.
 **
@@ -11,30 +11,28 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -44,15 +42,136 @@
 #include "qbluetoothsocket_p.h"
 #include "qbluetoothaddress.h"
 #include <QtCore/QLoggingCategory>
+#include <QtCore/QThread>
 #include <QtCore/QTime>
-#include <QtConcurrent/QtConcurrentRun>
+#include <QtCore/private/qjni_p.h>
 #include <QtAndroidExtras/QAndroidJniEnvironment>
 
-
 QT_BEGIN_NAMESPACE
 
 Q_DECLARE_LOGGING_CATEGORY(QT_BT_ANDROID)
 
+#define FALLBACK_CHANNEL 1
+#define USE_FALLBACK true
+
+Q_DECLARE_METATYPE(QAndroidJniObject)
+
+
+/* BluetoothSocket.connect() can block up to 10s. Therefore it must be
+ * in a separate thread. Unfortunately if BluetoothSocket.close() is
+ * called while connect() is still blocking the resulting behavior is not reliable.
+ * This may well be an Android platform bug. In any case, close() must
+ * be queued up until connect() has returned.
+ *
+ * The WorkerThread manages the connect() and close() calls. Interaction
+ * with the main thread happens via signals and slots. There is an accepted but
+ * undesirable side effect of this approach as the user may call connect()
+ * and close() and the socket would continue to successfully connect to
+ * the remote device just to immidiately close the physical connection again.
+ *
+ * WorkerThread and SocketConnectWorker are cleaned up via the threads
+ * finished() signal.
+ */
+
+class SocketConnectWorker : public QObject
+{
+    Q_OBJECT
+public:
+    SocketConnectWorker(const QAndroidJniObject& socket,
+                        const QAndroidJniObject& targetUuid)
+        : QObject(),
+          mSocketObject(socket),
+          mTargetUuid(targetUuid)
+    {
+        static int t = qRegisterMetaType<QAndroidJniObject>();
+        Q_UNUSED(t);
+    }
+
+signals:
+    void socketConnectDone(const QAndroidJniObject &socket);
+    void socketConnectFailed(const QAndroidJniObject &socket,
+                             const QAndroidJniObject &targetUuid);
+public slots:
+    void connectSocket()
+    {
+        QAndroidJniEnvironment env;
+
+        qCDebug(QT_BT_ANDROID) << "Connecting socket";
+        mSocketObject.callMethod<void>("connect");
+        if (env->ExceptionCheck()) {
+            env->ExceptionDescribe();
+            env->ExceptionClear();
+
+            emit socketConnectFailed(mSocketObject, mTargetUuid);
+            QThread::currentThread()->quit();
+            return;
+        }
+
+        qCDebug(QT_BT_ANDROID) << "Socket connection established";
+        emit socketConnectDone(mSocketObject);
+    }
+
+    void closeSocket()
+    {
+        qCDebug(QT_BT_ANDROID) << "Executing queued closeSocket()";
+
+        QAndroidJniEnvironment env;
+        mSocketObject.callMethod<void>("close");
+        if (env->ExceptionCheck()) {
+
+            qCWarning(QT_BT_ANDROID) << "Error during closure of abandoned socket";
+            env->ExceptionDescribe();
+            env->ExceptionClear();
+        }
+
+        QThread::currentThread()->quit();
+    }
+
+private:
+    QAndroidJniObject mSocketObject;
+    QAndroidJniObject mTargetUuid;
+};
+
+class WorkerThread: public QThread
+{
+    Q_OBJECT
+public:
+    WorkerThread()
+        : QThread(), workerPointer(0)
+    {
+    }
+
+    // Runs in same thread as QBluetoothSocketPrivate
+    void setupWorker(QBluetoothSocketPrivate* d_ptr, const QAndroidJniObject& socketObject,
+                     const QAndroidJniObject& uuidObject, bool useFallback)
+    {
+        SocketConnectWorker* worker = new SocketConnectWorker(
+                                            socketObject, uuidObject);
+        worker->moveToThread(this);
+
+        connect(this, &QThread::finished, worker, &QObject::deleteLater);
+        connect(this, &QThread::finished, this, &QObject::deleteLater);
+        connect(d_ptr, &QBluetoothSocketPrivate::connectJavaSocket,
+                worker, &SocketConnectWorker::connectSocket);
+        connect(d_ptr, &QBluetoothSocketPrivate::closeJavaSocket,
+                worker, &SocketConnectWorker::closeSocket);
+        connect(worker, &SocketConnectWorker::socketConnectDone,
+                d_ptr, &QBluetoothSocketPrivate::socketConnectSuccess);
+        if (useFallback) {
+            connect(worker, &SocketConnectWorker::socketConnectFailed,
+                    d_ptr, &QBluetoothSocketPrivate::fallbackSocketConnectFailed);
+        } else {
+            connect(worker, &SocketConnectWorker::socketConnectFailed,
+                d_ptr, &QBluetoothSocketPrivate::defaultSocketConnectFailed);
+        }
+
+        workerPointer = worker;
+    }
+
+private:
+    QPointer<SocketConnectWorker> workerPointer;
+};
+
 QBluetoothSocketPrivate::QBluetoothSocketPrivate()
   : socket(-1),
     socketType(QBluetoothServiceInfo::UnknownProtocol),
@@ -60,17 +179,20 @@ QBluetoothSocketPrivate::QBluetoothSocketPrivate()
     socketError(QBluetoothSocket::NoSocketError),
     connecting(false),
     discoveryAgent(0),
+    secFlags(QBluetooth::Secure),
     inputThread(0)
 {
     adapter = QAndroidJniObject::callStaticObjectMethod("android/bluetooth/BluetoothAdapter",
                                                         "getDefaultAdapter",
                                                         "()Landroid/bluetooth/BluetoothAdapter;");
-    qRegisterMetaType<QBluetoothSocket::SocketError>("QBluetoothSocket::SocketError");
-    qRegisterMetaType<QBluetoothSocket::SocketState>("QBluetoothSocket::SocketState");
+    qRegisterMetaType<QBluetoothSocket::SocketError>();
+    qRegisterMetaType<QBluetoothSocket::SocketState>();
 }
 
 QBluetoothSocketPrivate::~QBluetoothSocketPrivate()
 {
+    if (state != QBluetoothSocket::UnconnectedState)
+        emit closeJavaSocket();
 }
 
 bool QBluetoothSocketPrivate::ensureNativeSocket(QBluetoothServiceInfo::Protocol type)
@@ -82,49 +204,42 @@ bool QBluetoothSocketPrivate::ensureNativeSocket(QBluetoothServiceInfo::Protocol
     return false;
 }
 
-void QBluetoothSocketPrivate::connectToService(const QBluetoothAddress &address,
-                                               const QBluetoothUuid &uuid,
-                                               QIODevice::OpenMode openMode,
-                                               int fallbackServiceChannel)
-{
-    Q_Q(QBluetoothSocket);
-
-    q->setSocketState(QBluetoothSocket::ConnectingState);
-    QtConcurrent::run(this, &QBluetoothSocketPrivate::connectToServiceConc,
-                      address, uuid, openMode, fallbackServiceChannel);
-}
-
 bool QBluetoothSocketPrivate::fallBackConnect(QAndroidJniObject uuid, int channel)
 {
     qCWarning(QT_BT_ANDROID) << "Falling back to workaround.";
 
     QAndroidJniEnvironment env;
-    jclass remoteDeviceClazz = env->GetObjectClass(remoteDevice.object());
-    jmethodID getClassMethod = env->GetMethodID(remoteDeviceClazz, "getClass", "()Ljava/lang/Class;");
-    if (!getClassMethod) {
-        qCWarning(QT_BT_ANDROID) << "BluetoothDevice.getClass method could not be found.";
-        return false;
-    }
 
-
-    QAndroidJniObject remoteDeviceClass = QAndroidJniObject(env->CallObjectMethod(remoteDevice.object(), getClassMethod));
+    QAndroidJniObject remoteDeviceClass = remoteDevice.callObjectMethod("getClass", "()Ljava/lang/Class;");
     if (!remoteDeviceClass.isValid()) {
         qCWarning(QT_BT_ANDROID) << "Could not invoke BluetoothDevice.getClass.";
         return false;
     }
 
-    jclass classClass = env->FindClass("java/lang/Class");
-    jclass integerClass = env->FindClass("java/lang/Integer");
-    jfieldID integerType = env->GetStaticFieldID(integerClass, "TYPE", "Ljava/lang/Class;");
-    jobject integerObject = env->GetStaticObjectField(integerClass, integerType);
-    if (!integerObject) {
+    QAndroidJniObject integerObject = QAndroidJniObject::getStaticObjectField(
+                                            "java/lang/Integer", "TYPE", "Ljava/lang/Class;");
+    if (!integerObject.isValid()) {
         qCWarning(QT_BT_ANDROID) << "Could not get Integer.TYPE";
+        if (env->ExceptionCheck()) {
+            env->ExceptionDescribe();
+            env->ExceptionClear();
+        }
+
         return false;
     }
 
-    jobjectArray paramTypes = env->NewObjectArray(1, classClass, integerObject);
-    if (!paramTypes) {
+    jclass classClass = QJNIEnvironmentPrivate::findClass("java/lang/Class");
+    jobjectArray rawArray = env->NewObjectArray(1, classClass,
+                                                integerObject.object<jobject>());
+    QAndroidJniObject paramTypes(rawArray);
+    env->DeleteLocalRef(rawArray);
+    if (!paramTypes.isValid()) {
         qCWarning(QT_BT_ANDROID) << "Could not create new Class[]{Integer.TYPE}";
+
+        if (env->ExceptionCheck()) {
+            env->ExceptionDescribe();
+            env->ExceptionClear();
+        }
         return false;
     }
 
@@ -137,22 +252,34 @@ bool QBluetoothSocketPrivate::fallBackConnect(QAndroidJniObject uuid, int channe
         if (env->ExceptionCheck()) {
             env->ExceptionDescribe();
             env->ExceptionClear();
-        }
-
-        if (socketChannel
-                == remoteDevice.getStaticField<jint>("android/bluetooth/BluetoothDevice", "ERROR")) {
-            qCWarning(QT_BT_ANDROID) << "Cannot determine RFCOMM service channel.";
         } else {
-            qCWarning(QT_BT_ANDROID) << "Using found rfcomm channel" << socketChannel;
-            channel = socketChannel;
+            if (socketChannel
+                    == remoteDevice.getStaticField<jint>("android/bluetooth/BluetoothDevice", "ERROR")
+                || socketChannel == -1) {
+                qCWarning(QT_BT_ANDROID) << "Cannot determine RFCOMM service channel.";
+            } else {
+                qCWarning(QT_BT_ANDROID) << "Using found rfcomm channel" << socketChannel;
+                channel = socketChannel;
+            }
         }
     }
 
-    QAndroidJniObject method = remoteDeviceClass.callObjectMethod(
+    QAndroidJniObject method;
+    if (secFlags == QBluetooth::NoSecurity) {
+        qCDebug(QT_BT_ANDROID) << "Connnecting via insecure rfcomm";
+        method = remoteDeviceClass.callObjectMethod(
+                "getMethod",
+                "(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;",
+                QAndroidJniObject::fromString(QLatin1String("createInsecureRfcommSocket")).object<jstring>(),
+                paramTypes.object<jobjectArray>());
+    } else {
+        qCDebug(QT_BT_ANDROID) << "Connnecting via secure rfcomm";
+        method = remoteDeviceClass.callObjectMethod(
                 "getMethod",
                 "(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;",
                 QAndroidJniObject::fromString(QLatin1String("createRfcommSocket")).object<jstring>(),
-                paramTypes);
+                paramTypes.object<jobjectArray>());
+    }
     if (!method.isValid() || env->ExceptionCheck()) {
         qCWarning(QT_BT_ANDROID) << "Could not invoke getMethod";
         if (env->ExceptionCheck()) {
@@ -162,54 +289,63 @@ bool QBluetoothSocketPrivate::fallBackConnect(QAndroidJniObject uuid, int channe
         return false;
     }
 
-    jclass methodClass = env->GetObjectClass(method.object());
-    jmethodID invokeMethodId = env->GetMethodID(
-                methodClass, "invoke",
-                "(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;");
-    if (!invokeMethodId) {
-        qCWarning(QT_BT_ANDROID) << "Could not invoke method.";
-        return false;
-    }
+    jclass objectClass = QJNIEnvironmentPrivate::findClass("java/lang/Object");
+    QAndroidJniObject channelObject = QAndroidJniObject::callStaticObjectMethod(
+                                        "java/lang/Integer", "valueOf", "(I)Ljava/lang/Integer;", channel);
+    rawArray = env->NewObjectArray(1, objectClass, channelObject.object<jobject>());
 
-    jmethodID valueOfMethodId = env->GetStaticMethodID(integerClass, "valueOf", "(I)Ljava/lang/Integer;");
-    jclass objectClass = env->FindClass("java/lang/Object");
-    jobjectArray invokeParams = env->NewObjectArray(1, objectClass, env->CallStaticObjectMethod(integerClass, valueOfMethodId, channel));
-
-
-    jobject invokeResult = env->CallObjectMethod(method.object(), invokeMethodId,
-                                                 remoteDevice.object(), invokeParams);
-    if (!invokeResult)
+    QAndroidJniObject invokeResult = method.callObjectMethod("invoke",
+                                         "(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;",
+                                         remoteDevice.object<jobject>(), rawArray);
+    env->DeleteLocalRef(rawArray);
+    if (!invokeResult.isValid())
     {
         qCWarning(QT_BT_ANDROID) << "Invoke Resulted with error.";
         if (env->ExceptionCheck()) {
             env->ExceptionDescribe();
             env->ExceptionClear();
         }
+
         return false;
     }
 
     socketObject = QAndroidJniObject(invokeResult);
-    socketObject.callMethod<void>("connect");
-    if (env->ExceptionCheck()) {
-        env->ExceptionDescribe();
-        env->ExceptionClear();
 
-        qCWarning(QT_BT_ANDROID) << "Socket connect via workaround failed.";
+    WorkerThread *workerThread = new WorkerThread();
+    workerThread->setupWorker(this, socketObject, uuid, USE_FALLBACK);
+    workerThread->start();
+    emit connectJavaSocket();
 
-        return false;
-    }
-
-    qCWarning(QT_BT_ANDROID) << "Workaround invoked.";
+    qCWarning(QT_BT_ANDROID) << "Workaround thread invoked.";
     return true;
 }
 
-void QBluetoothSocketPrivate::connectToServiceConc(const QBluetoothAddress &address,
-                                                   const QBluetoothUuid &uuid, QIODevice::OpenMode openMode, int fallbackServiceChannel)
+
+/*
+ * The call order during a connectToService() is as follows:
+ *
+ * 1. call connectToService()
+ * 2. wait for execution of SocketConnectThread::run()
+ * 3. if threaded connect succeeds call socketConnectSuccess() via signals
+ *      -> done
+ * 4. if threaded connect fails call defaultSocketConnectFailed() via signals
+ * 5. call fallBackConnect()
+ * 6. if threaded connect on fallback channel succeeds call socketConnectSuccess()
+ *    via signals
+ *      -> done
+ * 7. if threaded connect on fallback channel fails call fallbackSocketConnectFailed()
+ *      -> complete failure of entire connectToService()
+ * */
+void QBluetoothSocketPrivate::connectToService(const QBluetoothAddress &address,
+                                               const QBluetoothUuid &uuid,
+                                               QIODevice::OpenMode openMode)
 {
     Q_Q(QBluetoothSocket);
     Q_UNUSED(openMode);
 
-    qCDebug(QT_BT_ANDROID) << "connectToServiceConc()" << address.toString() << uuid.toString();
+    qCDebug(QT_BT_ANDROID) << "connectToService()" << address.toString() << uuid.toString();
+
+    q->setSocketState(QBluetoothSocket::ConnectingState);
 
     if (!adapter.isValid()) {
         qCWarning(QT_BT_ANDROID) << "Device does not support Bluetooth";
@@ -253,9 +389,17 @@ void QBluetoothSocketPrivate::connectToServiceConc(const QBluetoothAddress &addr
                                                                        "(Ljava/lang/String;)Ljava/util/UUID;",
                                                                        inputString.object<jstring>());
 
-    socketObject = remoteDevice.callObjectMethod("createRfcommSocketToServiceRecord",
+    if (secFlags == QBluetooth::NoSecurity) {
+        qCDebug(QT_BT_ANDROID) << "Connnecting via insecure rfcomm";
+        socketObject = remoteDevice.callObjectMethod("createInsecureRfcommSocketToServiceRecord",
                                                  "(Ljava/util/UUID;)Landroid/bluetooth/BluetoothSocket;",
                                                  uuidObject.object<jobject>());
+    } else {
+        qCDebug(QT_BT_ANDROID) << "Connnecting via secure rfcomm";
+        socketObject = remoteDevice.callObjectMethod("createRfcommSocketToServiceRecord",
+                                                 "(Ljava/util/UUID;)Landroid/bluetooth/BluetoothSocket;",
+                                                 uuidObject.object<jobject>());
+    }
 
     if (env->ExceptionCheck()) {
         env->ExceptionDescribe();
@@ -269,22 +413,21 @@ void QBluetoothSocketPrivate::connectToServiceConc(const QBluetoothAddress &addr
         return;
     }
 
-    socketObject.callMethod<void>("connect");
-    if (env->ExceptionCheck()) {
-        env->ExceptionDescribe();
-        env->ExceptionClear();
+    WorkerThread *workerThread = new WorkerThread();
+    workerThread->setupWorker(this, socketObject, uuidObject, !USE_FALLBACK);
+    workerThread->start();
+    emit connectJavaSocket();
+}
 
-        bool success = fallBackConnect(uuidObject, fallbackServiceChannel);
-        if (!success) {
-            errorString = QBluetoothSocket::tr("Connection to service failed");
-            socketObject = remoteDevice = QAndroidJniObject();
-            q->setSocketError(QBluetoothSocket::ServiceNotFoundError);
-            q->setSocketState(QBluetoothSocket::UnconnectedState);
+void QBluetoothSocketPrivate::socketConnectSuccess(const QAndroidJniObject &socket)
+{
+    Q_Q(QBluetoothSocket);
+    QAndroidJniEnvironment env;
 
-            env->ExceptionClear(); //just in case
-            return;
-        }
-    }
+    // test we didn't get a success from a previous connect
+    // which was cleaned up late
+    if (socket != socketObject)
+        return;
 
     if (inputThread) {
         inputThread->deleteLater();
@@ -298,13 +441,7 @@ void QBluetoothSocketPrivate::connectToServiceConc(const QBluetoothAddress &addr
         env->ExceptionDescribe();
         env->ExceptionClear();
 
-        //close socket again
-        socketObject.callMethod<void>("close");
-        if (env->ExceptionCheck()) {
-            env->ExceptionDescribe();
-            env->ExceptionClear();
-        }
-
+        emit closeJavaSocket();
         socketObject = inputStream = outputStream = remoteDevice = QAndroidJniObject();
 
 
@@ -322,11 +459,7 @@ void QBluetoothSocketPrivate::connectToServiceConc(const QBluetoothAddress &addr
 
     if (!inputThread->run()) {
         //close socket again
-        socketObject.callMethod<void>("close");
-        if (env->ExceptionCheck()) {
-            env->ExceptionDescribe();
-            env->ExceptionClear();
-        }
+        emit closeJavaSocket();
 
         socketObject = inputStream = outputStream = remoteDevice = QAndroidJniObject();
 
@@ -339,16 +472,53 @@ void QBluetoothSocketPrivate::connectToServiceConc(const QBluetoothAddress &addr
         return;
     }
 
+    // only unbuffered behavior supported at this stage
+    q->setOpenMode(QIODevice::ReadWrite|QIODevice::Unbuffered);
+
     q->setSocketState(QBluetoothSocket::ConnectedState);
     emit q->connected();
 }
 
-void QBluetoothSocketPrivate::_q_writeNotify()
+void QBluetoothSocketPrivate::defaultSocketConnectFailed(
+        const QAndroidJniObject &socket, const QAndroidJniObject &targetUuid)
 {
+    Q_Q(QBluetoothSocket);
+
+    // test we didn't get a fail from a previous connect
+    // which was cleaned up late - should be same socket
+    if (socket != socketObject)
+        return;
+
+    bool success = fallBackConnect(targetUuid, FALLBACK_CHANNEL);
+    if (!success) {
+        errorString = QBluetoothSocket::tr("Connection to service failed");
+        socketObject = remoteDevice = QAndroidJniObject();
+        q->setSocketError(QBluetoothSocket::ServiceNotFoundError);
+        q->setSocketState(QBluetoothSocket::UnconnectedState);
+
+        QAndroidJniEnvironment env;
+        env->ExceptionClear(); // just in case
+        qCWarning(QT_BT_ANDROID) << "Workaround failed";
+    }
 }
 
-void QBluetoothSocketPrivate::_q_readNotify()
+void QBluetoothSocketPrivate::fallbackSocketConnectFailed(
+        const QAndroidJniObject &socket, const QAndroidJniObject &targetUuid)
 {
+    Q_UNUSED(targetUuid);
+    Q_Q(QBluetoothSocket);
+
+    // test we didn't get a fail from a previous connect
+    // which was cleaned up late - should be same socket
+    if (socket != socketObject)
+        return;
+
+    qCWarning(QT_BT_ANDROID) << "Socket connect via workaround failed.";
+    errorString = QBluetoothSocket::tr("Connection to service failed");
+    socketObject = remoteDevice = QAndroidJniObject();
+
+    q->setSocketError(QBluetoothSocket::ServiceNotFoundError);
+    q->setSocketState(QBluetoothSocket::UnconnectedState);
 }
 
 void QBluetoothSocketPrivate::abort()
@@ -370,23 +540,26 @@ void QBluetoothSocketPrivate::abort()
         if (inputThread)
             inputThread->prepareForClosure();
 
-        //triggers abort of input thread as well
-        socketObject.callMethod<void>("close");
-        if (env->ExceptionCheck()) {
+        emit closeJavaSocket();
 
-            qCWarning(QT_BT_ANDROID) << "Error during closure of socket";
-            env->ExceptionDescribe();
-            env->ExceptionClear();
-        }
+        inputStream = outputStream = socketObject = remoteDevice = QAndroidJniObject();
 
         if (inputThread) {
+            // inputThread exists hence we had a successful connect
+            // which means inputThread is responsible for setting Unconnected
+
             //don't delete here as signals caused by Java Thread are still
             //going to be emitted
             //delete occurs in inputThreadError()
             inputThread = 0;
+        } else {
+            // inputThread doesn't exist hence
+            // we abort in the middle of connect(). WorkerThread will do
+            // close() without further feedback. Therefore we have to set
+            // Unconnected (now) in advance
+            Q_Q(QBluetoothSocket);
+            q->setSocketState(QBluetoothSocket::UnconnectedState);
         }
-
-        inputStream = outputStream = socketObject = remoteDevice = QAndroidJniObject();
     }
 }
 
@@ -454,6 +627,7 @@ qint64 QBluetoothSocketPrivate::writeData(const char *data, qint64 maxSize)
     env->SetByteArrayRegion(nativeData, 0, (qint32)maxSize, reinterpret_cast<const jbyte*>(data));
     outputStream.callMethod<void>("write", "([BII)V", nativeData, 0, (qint32)maxSize);
     env->DeleteLocalRef(nativeData);
+    emit q->bytesWritten(maxSize);
 
     if (env->ExceptionCheck()) {
         qCWarning(QT_BT_ANDROID) << "Error while writing";
@@ -499,17 +673,17 @@ void QBluetoothSocketPrivate::inputThreadError(int errorCode)
         //cleanup internal objects
         //if it was call to local close()/abort() the objects are cleaned up already
 
-        QAndroidJniEnvironment env;
-        socketObject.callMethod<void>("close");
-        if (env->ExceptionCheck()) {
-            env->ExceptionDescribe();
-            env->ExceptionClear();
-        }
+        emit closeJavaSocket();
 
         inputStream = outputStream = remoteDevice = socketObject = QAndroidJniObject();
+        if (inputThread) {
+            // deleted already above (client->deleteLater())
+            inputThread = 0;
+        }
     }
 
     q->setSocketState(QBluetoothSocket::UnconnectedState);
+    q->setOpenMode(QIODevice::NotOpen);
     emit q->disconnected();
 }
 
@@ -585,7 +759,15 @@ bool QBluetoothSocketPrivate::setSocketDescriptor(const QAndroidJniObject &socke
 
 
     q->setSocketState(socketState);
-    q->setOpenMode(openMode);
+    q->setOpenMode(openMode | QIODevice::Unbuffered);
+
+    // WorkerThread manages all sockets for us
+    // When we come through here the socket was already connected by
+    // server socket listener (see QBluetoothServer)
+    // Therefore we only use WorkerThread to potentially close it later on
+    WorkerThread *workerThread = new WorkerThread();
+    workerThread->setupWorker(this, socketObject, QAndroidJniObject(), !USE_FALLBACK);
+    workerThread->start();
 
     if (openMode == QBluetoothSocket::ConnectedState)
         emit q->connected();
@@ -603,3 +785,5 @@ qint64 QBluetoothSocketPrivate::bytesAvailable() const
 }
 
 QT_END_NAMESPACE
+
+#include <qbluetoothsocket_android.moc>
