@@ -1,31 +1,37 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtBluetooth module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
-** As a special exception, The Qt Company gives you certain additional
-** rights. These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -42,6 +48,7 @@
 #include "bluez/bluez_data_p.h"
 
 #include <qplatformdefs.h>
+#include <QtCore/private/qcore_unix_p.h>
 
 #include <QtCore/QLoggingCategory>
 
@@ -64,6 +71,7 @@ QBluetoothSocketPrivate::QBluetoothSocketPrivate()
       connectWriteNotifier(0),
       connecting(false),
       discoveryAgent(0),
+      secFlags(QBluetooth::Authorization),
       lowEnergySocketType(0)
 {
 }
@@ -132,6 +140,27 @@ void QBluetoothSocketPrivate::connectToService(const QBluetoothAddress &address,
         return;
     }
 
+    // apply preferred security level
+    // ignore QBluetooth::Authentication -> not used anymore by kernel
+    struct bt_security security;
+    memset(&security, 0, sizeof(security));
+
+    if (secFlags & QBluetooth::Authorization)
+        security.level = BT_SECURITY_LOW;
+    if (secFlags & QBluetooth::Encryption)
+        security.level = BT_SECURITY_MEDIUM;
+    if (secFlags & QBluetooth::Secure)
+        security.level = BT_SECURITY_HIGH;
+
+    if (setsockopt(socket, SOL_BLUETOOTH, BT_SECURITY,
+                   &security, sizeof(security)) != 0) {
+        qCWarning(QT_BT_BLUEZ) << "Failed to set socket option, closing socket for safety" << errno;
+        qCWarning(QT_BT_BLUEZ) << "Error: " << qt_error_string(errno);
+        errorString = QBluetoothSocket::tr("Cannot set connection security level");
+        q->setSocketError(QBluetoothSocket::UnknownSocketError);
+        return;
+    }
+
     if (socketType == QBluetoothServiceInfo::RfcommProtocol) {
         sockaddr_rc addr;
 
@@ -213,13 +242,28 @@ void QBluetoothSocketPrivate::_q_writeNotify()
         char buf[1024];
 
         int size = txBuffer.read(buf, 1024);
-
-        if (::write(socket, buf, size) != size) {
-            errorString = QBluetoothSocket::tr("Network Error");
-            q->setSocketError(QBluetoothSocket::NetworkError);
-        }
-        else {
-            emit q->bytesWritten(size);
+        int writtenBytes = qt_safe_write(socket, buf, size);
+        if (writtenBytes < 0) {
+            switch (errno) {
+            case EAGAIN:
+                writtenBytes = 0;
+                txBuffer.ungetBlock(buf, size);
+                break;
+            default:
+                // every other case returns error
+                errorString = QBluetoothSocket::tr("Network Error: %1");
+                errorString.arg(qt_error_string(errno));
+                q->setSocketError(QBluetoothSocket::NetworkError);
+                break;
+            }
+        } else {
+            if (writtenBytes < size) {
+                // add remainder back to buffer
+                char* remainder = buf + writtenBytes;
+                txBuffer.ungetBlock(remainder, size - writtenBytes);
+            }
+            if (writtenBytes > 0)
+                emit q->bytesWritten(writtenBytes);
         }
 
         if (txBuffer.size()) {
@@ -245,9 +289,9 @@ void QBluetoothSocketPrivate::_q_readNotify()
         connectWriteNotifier->setEnabled(false);
         errorString = qt_error_string(errsv);
         qCWarning(QT_BT_BLUEZ) << Q_FUNC_INFO << socket << "error:" << readFromDevice << errorString;
-        if(errsv == EHOSTDOWN)
+        if (errsv == EHOSTDOWN)
             q->setSocketError(QBluetoothSocket::HostNotFoundError);
-        else
+        else if (errsv != ECONNRESET) // The other side closing the connection is not an error.
             q->setSocketError(QBluetoothSocket::UnknownSocketError);
 
         q->disconnectFromService();
@@ -287,20 +331,14 @@ QBluetoothAddress QBluetoothSocketPrivate::localAddress() const
         sockaddr_rc addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.rc_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.rc_bdaddr.b));
     } else if (socketType == QBluetoothServiceInfo::L2capProtocol) {
         sockaddr_l2 addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.l2_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getsockname(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.l2_bdaddr.b));
     }
 
     return QBluetoothAddress();
@@ -336,7 +374,7 @@ QString QBluetoothSocketPrivate::peerName() const
         if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) < 0)
             return QString();
 
-        convertAddress(addr.rc_bdaddr.b, bdaddr);
+        convertAddress(addr.rc_bdaddr.b, &bdaddr);
     } else if (socketType == QBluetoothServiceInfo::L2capProtocol) {
         sockaddr_l2 addr;
         socklen_t addrLength = sizeof(addr);
@@ -344,7 +382,7 @@ QString QBluetoothSocketPrivate::peerName() const
         if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) < 0)
             return QString();
 
-        convertAddress(addr.l2_bdaddr.b, bdaddr);
+        convertAddress(addr.l2_bdaddr.b, &bdaddr);
     } else {
         qCWarning(QT_BT_BLUEZ) << "peerName() called on socket of unknown type";
         return QString();
@@ -362,13 +400,17 @@ QString QBluetoothSocketPrivate::peerName() const
         if (reply.isError())
             return QString();
 
-        foreach (const QDBusObjectPath &path, reply.value().keys()) {
-            const InterfaceList ifaceList = reply.value().value(path);
-            foreach (const QString &iface, ifaceList.keys()) {
+        ManagedObjectList managedObjectList = reply.value();
+        for (ManagedObjectList::const_iterator it = managedObjectList.constBegin(); it != managedObjectList.constEnd(); ++it) {
+            const InterfaceList &ifaceList = it.value();
+
+            for (InterfaceList::const_iterator jt = ifaceList.constBegin(); jt != ifaceList.constEnd(); ++jt) {
+                const QString &iface = jt.key();
+                const QVariantMap &ifaceValues = jt.value();
+
                 if (iface == QStringLiteral("org.bluez.Device1")) {
-                    if (ifaceList.value(iface).value(QStringLiteral("Address")).toString()
-                            == peerAddress)
-                        return ifaceList.value(iface).value(QStringLiteral("Alias")).toString();
+                    if (ifaceValues.value(QStringLiteral("Address")).toString() == peerAddress)
+                        return ifaceValues.value(QStringLiteral("Alias")).toString();
                 }
             }
         }
@@ -385,13 +427,13 @@ QString QBluetoothSocketPrivate::peerName() const
         OrgBluezAdapterInterface adapter(QStringLiteral("org.bluez"), reply.value().path(),
                                          QDBusConnection::systemBus());
 
-        QDBusPendingReply<QDBusObjectPath> deviceObjectPath = adapter.CreateDevice(peerAddress);
+        QDBusPendingReply<QDBusObjectPath> deviceObjectPath = adapter.FindDevice(peerAddress);
         deviceObjectPath.waitForFinished();
         if (deviceObjectPath.isError()) {
-            if (deviceObjectPath.error().name() != QStringLiteral("org.bluez.Error.AlreadyExists"))
+            if (deviceObjectPath.error().name() != QStringLiteral("org.bluez.Error.DoesNotExist"))
                 return QString();
 
-            deviceObjectPath = adapter.FindDevice(peerAddress);
+            deviceObjectPath = adapter.CreateDevice(peerAddress);
             deviceObjectPath.waitForFinished();
             if (deviceObjectPath.isError())
                 return QString();
@@ -415,20 +457,14 @@ QBluetoothAddress QBluetoothSocketPrivate::peerAddress() const
         sockaddr_rc addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.rc_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.rc_bdaddr.b));
     } else if (socketType == QBluetoothServiceInfo::L2capProtocol) {
         sockaddr_l2 addr;
         socklen_t addrLength = sizeof(addr);
 
-        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0) {
-            quint64 bdaddr;
-            convertAddress(addr.l2_bdaddr.b, bdaddr);
-            return QBluetoothAddress(bdaddr);
-        }
+        if (::getpeername(socket, reinterpret_cast<sockaddr *>(&addr), &addrLength) == 0)
+            return QBluetoothAddress(convertAddress(addr.l2_bdaddr.b));
     }
 
     return QBluetoothAddress();
@@ -464,15 +500,23 @@ qint64 QBluetoothSocketPrivate::writeData(const char *data, qint64 maxSize)
     }
 
     if (q->openMode() & QIODevice::Unbuffered) {
-        if (::write(socket, data, maxSize) != maxSize) {
-            errorString = QBluetoothSocket::tr("Network Error");
-            q->setSocketError(QBluetoothSocket::NetworkError);
-            return -1;
+        int sz = ::qt_safe_write(socket, data, maxSize);
+        if (sz < 0) {
+            switch (errno) {
+            case EAGAIN:
+                sz = 0;
+                break;
+            default:
+                errorString = QBluetoothSocket::tr("Network Error: %1");
+                errorString.arg(qt_error_string(errno));
+                q->setSocketError(QBluetoothSocket::NetworkError);
+            }
         }
 
-        emit q->bytesWritten(maxSize);
+        if (sz > 0)
+            emit q->bytesWritten(sz);
 
-        return maxSize;
+        return sz;
     }
     else {
 
